@@ -20,13 +20,15 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
+
 #include <iconv.h>
 #include <errno.h>
+#include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 
 #include "glib.h"
-#include "config.h"
 
 #ifdef G_PLATFORM_WIN32
 #define STRICT
@@ -36,8 +38,15 @@
 
 #include "glibintl.h"
 
+#if defined(USE_LIBICONV) && !defined (_LIBICONV_H)
+#error libiconv in use but included iconv.h not from libiconv
+#endif
+#if !defined(USE_LIBICONV) && defined (_LIBICONV_H)
+#error libiconv not in use but included iconv.h is from libiconv
+#endif
+
 GQuark 
-g_convert_error_quark()
+g_convert_error_quark (void)
 {
   static GQuark quark;
   if (!quark)
@@ -46,13 +55,6 @@ g_convert_error_quark()
   return quark;
 }
 
-#if defined(USE_LIBICONV) && !defined (_LIBICONV_H)
-#error libiconv in use but included iconv.h not from libiconv
-#endif
-#if !defined(USE_LIBICONV) && defined (_LIBICONV_H)
-#error libiconv not in use but included iconv.h is from libiconv
-#endif
-
 static gboolean
 try_conversion (const char *to_codeset,
 		const char *from_codeset,
@@ -93,7 +95,7 @@ extern const char **_g_charset_get_aliases (const char *canonical_name);
  * @to_codeset: destination codeset
  * @from_codeset: source codeset
  * 
- * Same as the standard UNIX routine iconv_open(), but
+ * Same as the standard UNIX routine <function>iconv_open()</function>, but
  * may be implemented via libiconv on UNIX flavors that lack
  * a native implementation.
  * 
@@ -143,7 +145,7 @@ g_iconv_open (const gchar  *to_codeset,
  * @outbuf: converted output bytes
  * @outbytes_left: inout parameter, bytes available to fill in @outbuf
  * 
- * Same as the standard UNIX routine iconv(), but
+ * Same as the standard UNIX routine <function>iconv()</function>, but
  * may be implemented via libiconv on UNIX flavors that lack
  * a native implementation.
  *
@@ -168,10 +170,10 @@ g_iconv (GIConv   converter,
  * g_iconv_close:
  * @converter: a conversion descriptor from g_iconv_open()
  *
- * Same as the standard UNIX routine iconv_close(), but
+ * Same as the standard UNIX routine <function>iconv_close()</function>, but
  * may be implemented via libiconv on UNIX flavors that lack
  * a native implementation. Should be called to clean up
- * the conversion descriptor from iconv_open() when
+ * the conversion descriptor from g_iconv_open() when
  * you are done converting things.
  *
  * GLib provides g_convert() and g_locale_to_utf8() which are likely
@@ -187,30 +189,264 @@ g_iconv_close (GIConv converter)
   return iconv_close (cd);
 }
 
+
+#define ICONV_CACHE_SIZE   (16)
+
+struct _iconv_cache_bucket {
+  gchar *key;
+  guint32 refcount;
+  gboolean used;
+  iconv_t cd;
+};
+
+static GList *iconv_cache_list;
+static GHashTable *iconv_cache;
+static GHashTable *iconv_open_hash;
+static guint iconv_cache_size = 0;
+G_LOCK_DEFINE_STATIC (iconv_cache_lock);
+
+/* caller *must* hold the iconv_cache_lock */
+static void
+iconv_cache_init (void)
+{
+  static gboolean initialized = FALSE;
+  
+  if (initialized)
+    return;
+  
+  iconv_cache_list = NULL;
+  iconv_cache = g_hash_table_new (g_str_hash, g_str_equal);
+  iconv_open_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
+  
+  initialized = TRUE;
+}
+
+
+/**
+ * iconv_cache_bucket_new:
+ * @key: cache key
+ * @cd: iconv descriptor
+ *
+ * Creates a new cache bucket, inserts it into the cache and
+ * increments the cache size.
+ *
+ * Returns a pointer to the newly allocated cache bucket.
+ **/
+struct _iconv_cache_bucket *
+iconv_cache_bucket_new (const gchar *key, iconv_t cd)
+{
+  struct _iconv_cache_bucket *bucket;
+  
+  bucket = g_new (struct _iconv_cache_bucket, 1);
+  bucket->key = g_strdup (key);
+  bucket->refcount = 1;
+  bucket->used = TRUE;
+  bucket->cd = cd;
+  
+  g_hash_table_insert (iconv_cache, bucket->key, bucket);
+  
+  /* FIXME: if we sorted the list so items with few refcounts were
+     first, then we could expire them faster in iconv_cache_expire_unused () */
+  iconv_cache_list = g_list_prepend (iconv_cache_list, bucket);
+  
+  iconv_cache_size++;
+  
+  return bucket;
+}
+
+
+/**
+ * iconv_cache_bucket_expire:
+ * @node: cache bucket's node
+ * @bucket: cache bucket
+ *
+ * Expires a single cache bucket @bucket. This should only ever be
+ * called on a bucket that currently has no used iconv descriptors
+ * open.
+ *
+ * @node is not a required argument. If @node is not supplied, we
+ * search for it ourselves.
+ **/
+static void
+iconv_cache_bucket_expire (GList *node, struct _iconv_cache_bucket *bucket)
+{
+  g_hash_table_remove (iconv_cache, bucket->key);
+  
+  if (node == NULL)
+    node = g_list_find (iconv_cache_list, bucket);
+  
+  g_assert (node != NULL);
+  
+  if (node->prev)
+    {
+      node->prev->next = node->next;
+      if (node->next)
+        node->next->prev = node->prev;
+    }
+  else
+    {
+      iconv_cache_list = node->next;
+      if (node->next)
+        node->next->prev = NULL;
+    }
+  
+  g_list_free_1 (node);
+  
+  g_free (bucket->key);
+  g_iconv_close (bucket->cd);
+  g_free (bucket);
+  
+  iconv_cache_size--;
+}
+
+
+/**
+ * iconv_cache_expire_unused:
+ *
+ * Expires as many unused cache buckets as it needs to in order to get
+ * the total number of buckets < ICONV_CACHE_SIZE.
+ **/
+static void
+iconv_cache_expire_unused (void)
+{
+  struct _iconv_cache_bucket *bucket;
+  GList *node, *next;
+  
+  node = iconv_cache_list;
+  while (node && iconv_cache_size >= ICONV_CACHE_SIZE)
+    {
+      next = node->next;
+      
+      bucket = node->data;
+      if (bucket->refcount == 0)
+        iconv_cache_bucket_expire (node, bucket);
+      
+      node = next;
+    }
+}
+
 static GIConv
 open_converter (const gchar *to_codeset,
-                const gchar *from_codeset,
+		const gchar *from_codeset,
 		GError     **error)
 {
-  GIConv cd = g_iconv_open (to_codeset, from_codeset);
-
-  if (cd == (iconv_t) -1)
+  struct _iconv_cache_bucket *bucket;
+  gchar *key;
+  GIConv cd;
+  
+  /* create our key */
+  key = g_alloca (strlen (from_codeset) + strlen (to_codeset) + 2);
+  sprintf (key, "%s:%s", from_codeset, to_codeset);
+  
+  G_LOCK (iconv_cache_lock);
+  
+  /* make sure the cache has been initialized */
+  iconv_cache_init ();
+  
+  bucket = g_hash_table_lookup (iconv_cache, key);
+  if (bucket)
     {
-      /* Something went wrong.  */
-      if (errno == EINVAL)
-        g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_NO_CONVERSION,
-                     _("Conversion from character set '%s' to '%s' is not supported"),
-                     from_codeset, to_codeset);
+      if (bucket->used)
+        {
+          cd = g_iconv_open (to_codeset, from_codeset);
+          if (cd == (iconv_t) -1)
+            goto error;
+        }
       else
-        g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_FAILED,
-                     _("Could not open converter from '%s' to '%s': %s"),
-                     from_codeset, to_codeset, strerror (errno));
+        {
+          cd = bucket->cd;
+          bucket->used = TRUE;
+          
+          /* reset the descriptor */
+          g_iconv (cd, NULL, NULL, NULL, NULL);
+        }
+      
+      bucket->refcount++;
     }
-
+  else
+    {
+      cd = g_iconv_open (to_codeset, from_codeset);
+      if (cd == (iconv_t) -1)
+        goto error;
+      
+      iconv_cache_expire_unused ();
+      
+      bucket = iconv_cache_bucket_new (key, cd);
+    }
+  
+  g_hash_table_insert (iconv_open_hash, cd, bucket->key);
+  
+  G_UNLOCK (iconv_cache_lock);
+  
+  return cd;
+  
+ error:
+  
+  G_UNLOCK (iconv_cache_lock);
+  
+  /* Something went wrong.  */
+  if (errno == EINVAL)
+    g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_NO_CONVERSION,
+		 _("Conversion from character set '%s' to '%s' is not supported"),
+		 from_codeset, to_codeset);
+  else
+    g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_FAILED,
+		 _("Could not open converter from '%s' to '%s': %s"),
+		 from_codeset, to_codeset, strerror (errno));
+  
   return cd;
-
 }
 
+static int
+close_converter (GIConv converter)
+{
+  struct _iconv_cache_bucket *bucket;
+  const gchar *key;
+  iconv_t cd;
+  
+  cd = (iconv_t) converter;
+  
+  if (cd == (iconv_t) -1)
+    return 0;
+  
+  G_LOCK (iconv_cache_lock);
+  
+  key = g_hash_table_lookup (iconv_open_hash, cd);
+  if (key)
+    {
+      g_hash_table_remove (iconv_open_hash, cd);
+      
+      bucket = g_hash_table_lookup (iconv_cache, key);
+      g_assert (bucket);
+      
+      bucket->refcount--;
+      
+      if (cd == bucket->cd)
+        bucket->used = FALSE;
+      else
+        g_iconv_close (cd);
+      
+      if (!bucket->refcount && iconv_cache_size > ICONV_CACHE_SIZE)
+        {
+          /* expire this cache bucket */
+          iconv_cache_bucket_expire (NULL, bucket);
+        }
+    }
+  else
+    {
+      G_UNLOCK (iconv_cache_lock);
+      
+      g_warning ("This iconv context wasn't opened using open_converter");
+      
+      return g_iconv_close (converter);
+    }
+  
+  G_UNLOCK (iconv_cache_lock);
+  
+  return 0;
+}
+
+
 /**
  * g_convert:
  * @str:           the string to convert
@@ -226,11 +462,11 @@ open_converter (const gchar *to_codeset,
  *                 stored will the byte offset after the last valid
  *                 input sequence.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                 including the terminating NULL).
+ *                 including the terminating nul).
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  *
- * Convert a string from one character set to another.
+ * Converts a string from one character set to another.
  *
  * Return value: If the conversion was successful, a newly allocated
  *               nul-terminated string, which must be freed with
@@ -251,7 +487,7 @@ g_convert (const gchar *str,
   g_return_val_if_fail (str != NULL, NULL);
   g_return_val_if_fail (to_codeset != NULL, NULL);
   g_return_val_if_fail (from_codeset != NULL, NULL);
-     
+  
   cd = open_converter (to_codeset, from_codeset, error);
 
   if (cd == (GIConv) -1)
@@ -269,7 +505,7 @@ g_convert (const gchar *str,
 			      bytes_read, bytes_written,
 			      error);
   
-  g_iconv_close (cd);
+  close_converter (cd);
 
   return res;
 }
@@ -281,18 +517,18 @@ g_convert (const gchar *str,
  * @converter:     conversion descriptor from g_iconv_open()
  * @bytes_read:    location to store the number of bytes in the
  *                 input string that were successfully converted, or %NULL.
- *                 Even if the conversion was succesful, this may be 
+ *                 Even if the conversion was successful, this may be 
  *                 less than @len if there were partial characters
  *                 at the end of the input. If the error
  *                 #G_CONVERT_ERROR_ILLEGAL_SEQUENCE occurs, the value
  *                 stored will the byte offset after the last valid
  *                 input sequence.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                 including the terminating NULL).
+ *                 including the terminating nul).
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  *
- * Convert a string from one character set to another.
+ * Converts a string from one character set to another.
  *
  * Return value: If the conversion was successful, a newly allocated
  *               nul-terminated string, which must be freed with
@@ -407,21 +643,21 @@ g_convert_with_iconv (const gchar *str,
  *                as Unicode escapes \x{XXXX} or \x{XXXXXX}.
  * @bytes_read:   location to store the number of bytes in the
  *                input string that were successfully converted, or %NULL.
- *                Even if the conversion was succesful, this may be 
+ *                Even if the conversion was successful, this may be 
  *                less than @len if there were partial characters
  *                at the end of the input.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                including the terminating NULL).
+ *                including the terminating nul).
  * @error:        location to store the error occuring, or %NULL to ignore
  *                errors. Any of the errors in #GConvertError may occur.
  *
- * Convert a string from one character set to another, possibly
+ * Converts a string from one character set to another, possibly
  * including fallback sequences for characters not representable
  * in the output. Note that it is not guaranteed that the specification
  * for the fallback sequences in @fallback will be honored. Some
  * systems may do a approximate conversion from @from_codeset
- * to @to_codeset in their iconv() functions, in which case GLib
- * will simply return that approximate conversion.
+ * to @to_codeset in their <function>iconv()</function> functions, 
+ * in which case GLib will simply return that approximate conversion.
  *
  * Return value: If the conversion was successful, a newly allocated
  *               nul-terminated string, which must be freed with
@@ -497,7 +733,12 @@ g_convert_with_fallback (const gchar *str,
   utf8 = g_convert (str, len, "UTF-8", from_codeset, 
 		    bytes_read, &inbytes_remaining, error);
   if (!utf8)
-    return NULL;
+    {
+      close_converter (cd);
+      if (bytes_written)
+        *bytes_written = 0;
+      return NULL;
+    }
 
   /* Now the heart of the code. We loop through the UTF-8 string, and
    * whenever we hit an offending character, we form fallback, convert
@@ -594,10 +835,10 @@ g_convert_with_fallback (const gchar *str,
    */
   *outp = '\0';
   
-  g_iconv_close (cd);
+  close_converter (cd);
 
   if (bytes_written)
-    *bytes_written = outp - str;	/* Doesn't include '\0' */
+    *bytes_written = outp - dest;	/* Doesn't include '\0' */
 
   g_free (utf8);
 
@@ -618,6 +859,8 @@ g_convert_with_fallback (const gchar *str,
  * 
  */
 
+#ifndef G_PLATFORM_WIN32
+
 static gchar *
 strdup_len (const gchar *string,
 	    gssize       len,
@@ -658,6 +901,8 @@ strdup_len (const gchar *string,
   return g_strndup (string, real_len);
 }
 
+#endif
+
 /**
  * g_locale_to_utf8:
  * @opsysstring:   a string in the encoding of the current locale
@@ -665,14 +910,14 @@ strdup_len (const gchar *string,
  *                 nul-terminated.
  * @bytes_read:    location to store the number of bytes in the
  *                 input string that were successfully converted, or %NULL.
- *                 Even if the conversion was succesful, this may be 
+ *                 Even if the conversion was successful, this may be 
  *                 less than @len if there were partial characters
  *                 at the end of the input. If the error
  *                 #G_CONVERT_ERROR_ILLEGAL_SEQUENCE occurs, the value
  *                 stored will the byte offset after the last valid
  *                 input sequence.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                 including the terminating NULL).
+ *                 including the terminating nul).
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  * 
@@ -805,14 +1050,14 @@ g_locale_to_utf8 (const gchar  *opsysstring,
  *                 nul-terminated.
  * @bytes_read:    location to store the number of bytes in the
  *                 input string that were successfully converted, or %NULL.
- *                 Even if the conversion was succesful, this may be 
+ *                 Even if the conversion was successful, this may be 
  *                 less than @len if there were partial characters
  *                 at the end of the input. If the error
  *                 #G_CONVERT_ERROR_ILLEGAL_SEQUENCE occurs, the value
  *                 stored will the byte offset after the last valid
  *                 input sequence.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                 including the terminating NULL).
+ *                 including the terminating nul).
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  * 
@@ -951,14 +1196,14 @@ g_locale_from_utf8 (const gchar *utf8string,
  *                 nul-terminated.
  * @bytes_read:    location to store the number of bytes in the
  *                 input string that were successfully converted, or %NULL.
- *                 Even if the conversion was succesful, this may be 
+ *                 Even if the conversion was successful, this may be 
  *                 less than @len if there were partial characters
  *                 at the end of the input. If the error
  *                 #G_CONVERT_ERROR_ILLEGAL_SEQUENCE occurs, the value
  *                 stored will the byte offset after the last valid
  *                 input sequence.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                 including the terminating NULL).
+ *                 including the terminating nul).
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  * 
@@ -996,14 +1241,14 @@ g_filename_to_utf8 (const gchar *opsysstring,
  *                 nul-terminated.
  * @bytes_read:    location to store the number of bytes in the
  *                 input string that were successfully converted, or %NULL.
- *                 Even if the conversion was succesful, this may be 
+ *                 Even if the conversion was successful, this may be 
  *                 less than @len if there were partial characters
  *                 at the end of the input. If the error
  *                 #G_CONVERT_ERROR_ILLEGAL_SEQUENCE occurs, the value
  *                 stored will the byte offset after the last valid
  *                 input sequence.
  * @bytes_written: the number of bytes stored in the output buffer (not 
- *                 including the terminating NULL).
+ *                 including the terminating nul).
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  * 
@@ -1064,13 +1309,19 @@ typedef enum {
 } UnsafeCharacterSet;
 
 static const guchar acceptable[96] = {
- /* X0   X1   X2   X3   X4   X5   X6   X7   X8   X9   XA   XB   XC   XD   XE   XF */
-  0x00,0x3F,0x20,0x20,0x20,0x00,0x2C,0x3F,0x3F,0x3F,0x3F,0x22,0x20,0x3F,0x3F,0x1C, /* 2X  !"#$%&'()*+,-./   */
-  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x38,0x20,0x20,0x2C,0x20,0x2C, /* 3X 0123456789:;<=>?   */
-  0x30,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F, /* 4X @ABCDEFGHIJKLMNO   */
-  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x20,0x3F, /* 5X PQRSTUVWXYZ[\]^_   */
-  0x20,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F, /* 6X `abcdefghijklmno   */
-  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x3F,0x20  /* 7X pqrstuvwxyz{|}~DEL */
+  /* A table of the ASCII chars from space (32) to DEL (127) */
+  /*      !    "    #    $    %    &    '    (    )    *    +    ,    -    .    / */ 
+  0x00,0x3F,0x20,0x20,0x20,0x00,0x2C,0x3F,0x3F,0x3F,0x3F,0x22,0x20,0x3F,0x3F,0x1C,
+  /* 0    1    2    3    4    5    6    7    8    9    :    ;    <    =    >    ? */
+  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x38,0x20,0x20,0x2C,0x20,0x2C,
+  /* @    A    B    C    D    E    F    G    H    I    J    K    L    M    N    O */
+  0x30,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,
+  /* P    Q    R    S    T    U    V    W    X    Y    Z    [    \    ]    ^    _ */
+  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x20,0x3F,
+  /* `    a    b    c    d    e    f    g    h    i    j    k    l    m    n    o */
+  0x20,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,
+  /* p    q    r    s    t    u    v    w    x    y    z    {    |    }    ~  DEL */
+  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x3F,0x20
 };
 
 static const gchar hex[16] = "0123456789ABCDEF";
@@ -1101,7 +1352,7 @@ g_escape_uri_string (const gchar *string,
   use_mask = mask;
   for (p = string; *p != '\0'; p++)
     {
-      c = *p;
+      c = (guchar) *p;
       if (!ACCEPTABLE (c)) 
 	unacceptable++;
     }
@@ -1111,7 +1362,7 @@ g_escape_uri_string (const gchar *string,
   use_mask = mask;
   for (q = result, p = string; *p != '\0'; p++)
     {
-      c = (unsigned char)*p;
+      c = (guchar) *p;
       
       if (!ACCEPTABLE (c))
 	{
@@ -1137,6 +1388,23 @@ g_escape_file_uri (const gchar *hostname,
   char *escaped_path;
   char *res;
 
+#ifdef G_OS_WIN32
+  char *p, *backslash;
+
+  /* Turn backslashes into forward slashes. That's what Netscape
+   * does, and they are actually more or less equivalent in Windows.
+   */
+  
+  pathname = g_strdup (pathname);
+  p = (char *) pathname;
+  
+  while ((backslash = strchr (p, '\\')) != NULL)
+    {
+      *backslash = '/';
+      p = backslash + 1;
+    }
+#endif
+
   if (hostname && *hostname != '\0')
     {
       escaped_hostname = g_escape_uri_string (hostname, UNSAFE_HOST);
@@ -1150,6 +1418,10 @@ g_escape_file_uri (const gchar *hostname,
 		     escaped_path,
 		     NULL);
 
+#ifdef G_OS_WIN32
+  g_free ((char *) pathname);
+#endif
+
   g_free (escaped_hostname);
   g_free (escaped_path);
   
@@ -1162,12 +1434,11 @@ unescape_character (const char *scanner)
   int first_digit;
   int second_digit;
 
-  first_digit = g_ascii_xdigit_value (*scanner++);
-  
+  first_digit = g_ascii_xdigit_value (scanner[0]);
   if (first_digit < 0) 
     return -1;
   
-  second_digit = g_ascii_xdigit_value (*scanner++);
+  second_digit = g_ascii_xdigit_value (scanner[1]);
   if (second_digit < 0) 
     return -1;
   
@@ -1175,13 +1446,14 @@ unescape_character (const char *scanner)
 }
 
 static gchar *
-g_unescape_uri_string (const gchar *escaped,
-		       const gchar *illegal_characters,
-		       int          len)
+g_unescape_uri_string (const char *escaped,
+		       int         len,
+		       const char *illegal_escaped_characters,
+		       gboolean    ascii_must_not_be_escaped)
 {
   const gchar *in, *in_end;
   gchar *out, *result;
-  int character;
+  int c;
   
   if (escaped == NULL)
     return NULL;
@@ -1189,42 +1461,99 @@ g_unescape_uri_string (const gchar *escaped,
   if (len < 0)
     len = strlen (escaped);
 
-    result = g_malloc (len + 1);
+  result = g_malloc (len + 1);
   
   out = result;
-  for (in = escaped, in_end = escaped + len; in < in_end && *in != '\0'; in++)
+  for (in = escaped, in_end = escaped + len; in < in_end; in++)
     {
-      character = *in;
-      if (character == '%')
+      c = *in;
+
+      if (c == '%')
 	{
-	  character = unescape_character (in + 1);
-      
-	  /* Check for an illegal character. We consider '\0' illegal here. */
-	  if (character == 0
-	      || (illegal_characters != NULL
-		  && strchr (illegal_characters, (char)character) != NULL))
-	    {
-	      g_free (result);
-	      return NULL;
-	    }
+	  /* catch partial escape sequences past the end of the substring */
+	  if (in + 3 > in_end)
+	    break;
+
+	  c = unescape_character (in + 1);
+
+	  /* catch bad escape sequences and NUL characters */
+	  if (c <= 0)
+	    break;
+
+	  /* catch escaped ASCII */
+	  if (ascii_must_not_be_escaped && c <= 0x7F)
+	    break;
+
+	  /* catch other illegal escaped characters */
+	  if (strchr (illegal_escaped_characters, c) != NULL)
+	    break;
+
 	  in += 2;
 	}
-      *out++ = character;
+
+      *out++ = c;
     }
   
+  g_assert (out - result <= len);
   *out = '\0';
-  
-  g_assert (out - result <= strlen (escaped));
 
-  if (!g_utf8_validate (result, -1, NULL))
+  if (in != in_end || !g_utf8_validate (result, -1, NULL))
     {
       g_free (result);
       return NULL;
     }
-  
+
   return result;
 }
 
+static gboolean
+is_escalphanum (gunichar c)
+{
+  return c > 0x7F || g_ascii_isalnum (c);
+}
+
+static gboolean
+is_escalpha (gunichar c)
+{
+  return c > 0x7F || g_ascii_isalpha (c);
+}
+
+/* allows an empty string */
+static gboolean
+hostname_validate (const char *hostname)
+{
+  const char *p;
+  gunichar c, first_char, last_char;
+
+  p = hostname;
+  if (*p == '\0')
+    return TRUE;
+  do
+    {
+      /* read in a label */
+      c = g_utf8_get_char (p);
+      p = g_utf8_next_char (p);
+      if (!is_escalphanum (c))
+	return FALSE;
+      first_char = c;
+      do
+	{
+	  last_char = c;
+	  c = g_utf8_get_char (p);
+	  p = g_utf8_next_char (p);
+	}
+      while (is_escalphanum (c) || c == '-');
+      if (last_char == '-')
+	return FALSE;
+      
+      /* if that was the last label, check that it was a toplabel */
+      if (c == '\0' || (c == '.' && *p == '\0'))
+	return is_escalpha (first_char);
+    }
+  while (c == '.');
+  return FALSE;
+}
+
 /**
  * g_filename_from_uri:
  * @uri: a uri describing a filename (escaped, encoded in UTF-8).
@@ -1237,7 +1566,7 @@ g_unescape_uri_string (const gchar *escaped,
  * Converts an escaped UTF-8 encoded URI to a local filename in the
  * encoding used for filenames. 
  * 
- * Return value: a newly allocated string holding the resulting
+ * Return value: a newly-allocated string holding the resulting
  *               filename, or %NULL on an error.
  **/
 gchar *
@@ -1251,6 +1580,9 @@ g_filename_from_uri (const char *uri,
   char *result;
   char *filename;
   int offs;
+#ifdef G_OS_WIN32
+  char *p, *slash;
+#endif
 
   if (hostname)
     *hostname = NULL;
@@ -1290,11 +1622,14 @@ g_filename_from_uri (const char *uri,
 	  return NULL;
 	}
 
-      unescaped_hostname = g_unescape_uri_string (host_part, "", path_part - host_part);
-      if (unescaped_hostname == NULL)
+      unescaped_hostname = g_unescape_uri_string (host_part, path_part - host_part, "", TRUE);
+
+      if (unescaped_hostname == NULL ||
+	  !hostname_validate (unescaped_hostname))
 	{
+	  g_free (unescaped_hostname);
 	  g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_BAD_URI,
-		       _("The hostname of the URI '%s' contains invalidly escaped characters"),
+		       _("The hostname of the URI '%s' is invalid"),
 		       uri);
 	  return NULL;
 	}
@@ -1305,7 +1640,7 @@ g_filename_from_uri (const char *uri,
 	g_free (unescaped_hostname);
     }
 
-  filename = g_unescape_uri_string (path_part, "/", -1);
+  filename = g_unescape_uri_string (path_part, -1, "/", FALSE);
 
   if (filename == NULL)
     {
@@ -1315,11 +1650,39 @@ g_filename_from_uri (const char *uri,
       return NULL;
     }
 
-  /* DOS uri's are like "file://host/c:\foo", so we need to check if we need to
-   * drop the initial slash */
   offs = 0;
-  if (g_path_is_absolute (filename+1))
-    offs = 1;
+#ifdef G_OS_WIN32
+  /* Drop localhost */
+  if (hostname && *hostname != NULL &&
+      g_ascii_strcasecmp (*hostname, "localhost") == 0)
+    {
+      g_free (*hostname);
+      *hostname = NULL;
+    }
+
+  /* Turn slashes into backslashes, because that's the canonical spelling */
+  p = filename;
+  while ((slash = strchr (p, '/')) != NULL)
+    {
+      *slash = '\\';
+      p = slash + 1;
+    }
+
+  /* Windows URIs with a drive letter can be like "file://host/c:/foo"
+   * or "file://host/c|/foo" (some Netscape versions). In those cases, start
+   * the filename from the drive letter.
+   */
+  if (g_ascii_isalpha (filename[1]))
+    {
+      if (filename[2] == ':')
+	offs = 1;
+      else if (filename[2] == '|')
+	{
+	  filename[2] = ':';
+	  offs = 1;
+	}
+    }
+#endif
   
   result = g_filename_from_utf8 (filename + offs, -1, NULL, NULL, error);
   g_free (filename);
@@ -1337,12 +1700,12 @@ g_filename_from_uri (const char *uri,
  * 
  * Converts an absolute filename to an escaped UTF-8 encoded URI.
  * 
- * Return value: a newly allocated string holding the resulting
+ * Return value: a newly-allocated string holding the resulting
  *               URI, or %NULL on an error.
  **/
 gchar *
 g_filename_to_uri   (const char *filename,
-		     char       *hostname,
+		     const char *hostname,
 		     GError    **error)
 {
   char *escaped_uri;
@@ -1358,19 +1721,25 @@ g_filename_to_uri   (const char *filename,
       return NULL;
     }
 
+  if (hostname &&
+      !(g_utf8_validate (hostname, -1, NULL)
+	&& hostname_validate (hostname)))
+    {
+      g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_ILLEGAL_SEQUENCE,
+		   _("Invalid hostname"));
+      return NULL;
+    }
+  
   utf8_filename = g_filename_to_utf8 (filename, -1, NULL, NULL, error);
   if (utf8_filename == NULL)
     return NULL;
   
-  if (hostname &&
-      !g_utf8_validate (hostname, -1, NULL))
-    {
-      g_free (utf8_filename);
-      g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_ILLEGAL_SEQUENCE,
-		   _("Invalid byte sequence in hostname"));
-      return NULL;
-    }
-  
+#ifdef G_OS_WIN32
+  /* Don't use localhost unnecessarily */
+  if (hostname && g_ascii_strcasecmp (hostname, "localhost") == 0)
+    hostname = NULL;
+#endif
+
   escaped_uri = g_escape_file_uri (hostname,
 				   utf8_filename);
   g_free (utf8_filename);
