@@ -28,6 +28,7 @@
 #endif
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <string.h>
 #include <errno.h>
 #include <sys/types.h>
@@ -69,42 +70,59 @@
  * @filename: a filename to test
  * @test: bitfield of #GFileTest flags
  * 
- * Returns TRUE if any of the tests in the bitfield @test are
- * TRUE. For example, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)
- * will return TRUE if the file exists; the check whether it's
- * a directory doesn't matter since the existence test is TRUE.
- * With the current set of available tests, there's no point
+ * Returns %TRUE if any of the tests in the bitfield @test are
+ * %TRUE. For example, <literal>(G_FILE_TEST_EXISTS | 
+ * G_FILE_TEST_IS_DIR)</literal> will return %TRUE if the file exists; 
+ * the check whether it's a directory doesn't matter since the existence 
+ * test is %TRUE. With the current set of available tests, there's no point
  * passing in more than one test at a time.
+ * 
+ * Apart from #G_FILE_TEST_IS_SYMLINK all tests follow symbolic links,
+ * so for a symbolic link to a regular file g_file_test() will return
+ * %TRUE for both #G_FILE_TEST_IS_SYMLINK and #G_FILE_TEST_IS_REGULAR.
  *
- * Return value: whether a test was TRUE
+ * Note, that for a dangling symbolic link g_file_test() will return
+ * %TRUE for #G_FILE_TEST_IS_SYMLINK and %FALSE for all other flags.
+ *
+ * Return value: whether a test was %TRUE
  **/
 gboolean
 g_file_test (const gchar *filename,
              GFileTest    test)
 {
-  if (test & G_FILE_TEST_EXISTS)
-    return (access (filename, F_OK) == 0);
-  else if (test & G_FILE_TEST_IS_EXECUTABLE)
-    return (access (filename, X_OK) == 0);
-  else
+  if ((test & G_FILE_TEST_EXISTS) && (access (filename, F_OK) == 0))
+    return TRUE;
+  
+  if ((test & G_FILE_TEST_IS_EXECUTABLE) && (access (filename, X_OK) == 0))
+    return TRUE;
+
+  if (test & G_FILE_TEST_IS_SYMLINK)
+    {
+#ifdef G_OS_WIN32
+      /* no sym links on win32, no lstat in msvcrt */
+#else
+      struct stat s;
+
+      if ((lstat (filename, &s) == 0) && S_ISLNK (s.st_mode))
+        return TRUE;
+#endif
+    }
+  
+  if (test & (G_FILE_TEST_IS_REGULAR | G_FILE_TEST_IS_DIR))
     {
       struct stat s;
       
-      if (stat (filename, &s) < 0)
-        return FALSE;
-
-      if ((test & G_FILE_TEST_IS_REGULAR) &&
-          S_ISREG (s.st_mode))
-        return TRUE;
-      else if ((test & G_FILE_TEST_IS_DIR) &&
-               S_ISDIR (s.st_mode))
-        return TRUE;
-      else if ((test & G_FILE_TEST_IS_SYMLINK) &&
-               S_ISLNK (s.st_mode))
-        return TRUE;
-      else
-        return FALSE;
+      if (stat (filename, &s) == 0)
+	{
+	  if ((test & G_FILE_TEST_IS_REGULAR) && S_ISREG (s.st_mode))
+	    return TRUE;
+	  
+	  if ((test & G_FILE_TEST_IS_DIR) && S_ISDIR (s.st_mode))
+	    return TRUE;
+	}
     }
+
+  return FALSE;
 }
 
 GQuark
@@ -121,16 +139,16 @@ g_file_error_quark (void)
  * g_file_error_from_errno:
  * @err_no: an "errno" value
  * 
- * Gets a #GFileError constant based on the passed-in errno.
- * For example, if you pass in EEXIST this function returns
- * #G_FILE_ERROR_EXIST. Unlike errno values, you can portably
+ * Gets a #GFileError constant based on the passed-in @errno.
+ * For example, if you pass in %EEXIST this function returns
+ * #G_FILE_ERROR_EXIST. Unlike @errno values, you can portably
  * assume that all #GFileError values will exist.
  *
  * Normally a #GFileError value goes into a #GError returned
  * from a function that manipulates files. So you would use
  * g_file_error_from_errno() when constructing a #GError.
  * 
- * Return value: #GFileError corresponding to the given errno
+ * Return value: #GFileError corresponding to the given @errno
  **/
 GFileError
 g_file_error_from_errno (gint err_no)
@@ -290,15 +308,37 @@ get_contents_stdio (const gchar *filename,
 {
   gchar buf[2048];
   size_t bytes;
-  GString *str;
-
-  g_assert (f != NULL);
+  char *str;
+  size_t total_bytes;
+  size_t total_allocated;
   
-  str = g_string_new ("");
+  g_assert (f != NULL);
+
+#define STARTING_ALLOC 64
+  
+  total_bytes = 0;
+  total_allocated = STARTING_ALLOC;
+  str = g_malloc (STARTING_ALLOC);
   
   while (!feof (f))
     {
       bytes = fread (buf, 1, 2048, f);
+
+      while ((total_bytes + bytes + 1) > total_allocated)
+        {
+          total_allocated *= 2;
+          str = g_try_realloc (str, total_allocated);
+
+          if (str == NULL)
+            {
+              g_set_error (error,
+                           G_FILE_ERROR,
+                           G_FILE_ERROR_NOMEM,
+                           _("Could not allocate %lu bytes to read file \"%s\""),
+                           (gulong) total_allocated, filename);
+              goto error;
+            }
+        }
       
       if (ferror (f))
         {
@@ -308,23 +348,30 @@ get_contents_stdio (const gchar *filename,
                        _("Error reading file '%s': %s"),
                        filename, strerror (errno));
 
-          g_string_free (str, TRUE);
-	  fclose (f);
-          
-          return FALSE;
+          goto error;
         }
 
-      g_string_append_len (str, buf, bytes);
+      memcpy (str + total_bytes, buf, bytes);
+      total_bytes += bytes;
     }
 
   fclose (f);
 
-  if (length)
-    *length = str->len;
+  str[total_bytes] = '\0';
   
-  *contents = g_string_free (str, FALSE);
+  if (length)
+    *length = total_bytes;
+  
+  *contents = str;
+  
+  return TRUE;
 
-  return TRUE;  
+ error:
+
+  g_free (str);
+  fclose (f);
+  
+  return FALSE;  
 }
 
 #ifndef G_OS_WIN32
@@ -340,11 +387,24 @@ get_contents_regfile (const gchar *filename,
   gchar *buf;
   size_t bytes_read;
   size_t size;
-      
+  size_t alloc_size;
+  
   size = stat_buf->st_size;
 
-  buf = g_new (gchar, size + 1);
-      
+  alloc_size = size + 1;
+  buf = g_try_malloc (alloc_size);
+
+  if (buf == NULL)
+    {
+      g_set_error (error,
+                   G_FILE_ERROR,
+                   G_FILE_ERROR_NOMEM,
+                   _("Could not allocate %lu bytes to read file \"%s\""),
+                   (gulong) alloc_size, filename);
+
+      return FALSE;
+    }
+  
   bytes_read = 0;
   while (bytes_read < size)
     {
@@ -489,17 +549,14 @@ get_contents_win32 (const gchar *filename,
  * @error: return location for a #GError
  * 
  * Reads an entire file into allocated memory, with good error
- * checking. If @error is set, FALSE is returned, and @contents is set
- * to NULL. If TRUE is returned, @error will not be set, and @contents
+ * checking. If @error is set, %FALSE is returned, and @contents is set
+ * to %NULL. If %TRUE is returned, @error will not be set, and @contents
  * will be set to the file contents.  The string stored in @contents
- * will be nul-terminated, so for text files you can pass NULL for the
+ * will be nul-terminated, so for text files you can pass %NULL for the
  * @length argument.  The error domain is #G_FILE_ERROR. Possible
  * error codes are those in the #GFileError enumeration.
  *
- * FIXME currently crashes if the file is too big to fit in memory;
- * should probably use g_try_malloc() when we have that function.
- * 
- * Return value: TRUE on success, FALSE if error is set
+ * Return value: %TRUE on success, %FALSE if error is set
  **/
 gboolean
 g_file_get_contents (const gchar *filename,
@@ -529,18 +586,19 @@ g_file_get_contents (const gchar *filename,
  * g_mkstemp:
  * @tmpl: template filename
  *
- * Open a temporary file. See "man mkstemp" on most UNIX-like systems.
- * This is a portability wrapper, which simply calls mkstemp() on systems
- * that have it, and implements it in GLib otherwise.
+ * Opens a temporary file. See the <function>mkstemp()</function> documentation
+ * on most UNIX-like systems. This is a portability wrapper, which simply calls 
+ * <function>mkstemp()</function> on systems that have it, and implements 
+ * it in GLib otherwise.
  *
- * The parameter is a string that should match the rules for mktemp, i.e.
- * end in "XXXXXX". The X string will be modified to form the name
- * of a file that didn't exist.
+ * The parameter is a string that should match the rules for
+ * <function>mkstemp()</function>, i.e. end in "XXXXXX". The X string will 
+ * be modified to form the name of a file that didn't exist.
  *
- * Return value: A file handle (as from open()) to the file
+ * Return value: A file handle (as from <function>open()</function>) to the file
  * opened for reading and writing. The file is opened in binary mode
  * on platforms where there is a difference. The file handle should be
- * closed with close(). In case of errors, -1 is returned.
+ * closed with <function>close()</function>. In case of errors, -1 is returned.
  */
 int
 g_mkstemp (char *tmpl)
@@ -604,7 +662,7 @@ g_mkstemp (char *tmpl)
 
 /**
  * g_file_open_tmp:
- * @tmpl: Template for file name, as in g_mkstemp, basename only
+ * @tmpl: Template for file name, as in g_mkstemp(), basename only
  * @name_used: location to store actual name used
  * @error: return location for a #GError
  *
@@ -612,22 +670,22 @@ g_mkstemp (char *tmpl)
  * files (as returned by g_get_tmp_dir()). 
  *
  * @tmpl should be a string ending with six 'X' characters, as the
- * parameter to g_mkstemp() (or mkstemp()). However, unlike these
- * functions, the template should only be a basename, no directory
- * components are allowed. If template is NULL, a default template is
- * used.
+ * parameter to g_mkstemp() (or <function>mkstemp()</function>). 
+ * However, unlike these functions, the template should only be a 
+ * basename, no directory components are allowed. If template is %NULL, 
+ * a default template is used.
  *
- * Note that in contrast to g_mkstemp() (and mkstemp()) @tmpl is not
- * modified, and might thus be a read-only literal string.
+ * Note that in contrast to g_mkstemp() (and <function>mkstemp()</function>) 
+ * @tmpl is not modified, and might thus be a read-only literal string.
  *
- * The actual name used is returned in @name_used if non-NULL. This
- * string should be freed with g_free when not needed any longer.
+ * The actual name used is returned in @name_used if non-%NULL. This
+ * string should be freed with g_free() when not needed any longer.
  *
- * Return value: A file handle (as from open()) to the file
- * opened for reading and writing. The file is opened in binary mode
- * on platforms where there is a difference. The file handle should be
- * closed with close(). In case of errors, -1 is returned and
- * @error will be set.
+ * Return value: A file handle (as from <function>open()</function>) to 
+ * the file opened for reading and writing. The file is opened in binary 
+ * mode on platforms where there is a difference. The file handle should be
+ * closed with <function>close()</function>. In case of errors, -1 is returned 
+ * and @error will be set.
  **/
 int
 g_file_open_tmp (const char *tmpl,
@@ -651,7 +709,7 @@ g_file_open_tmp (const char *tmpl,
       g_set_error (error,
 		   G_FILE_ERROR,
 		   G_FILE_ERROR_FAILED,
-		   _("Template '%s' illegal, should not contain a '%s'"),
+		   _("Template '%s' invalid, should not contain a '%s'"),
 		   tmpl, G_DIR_SEPARATOR_S);
 
       return -1;
@@ -697,3 +755,123 @@ g_file_open_tmp (const char *tmpl,
 
   return retval;
 }
+
+static gchar *
+g_build_pathv (const gchar *separator,
+	       const gchar *first_element,
+	       va_list      args)
+{
+  GString *result;
+  gint separator_len = strlen (separator);
+  gboolean is_first = TRUE;
+  const gchar *next_element;
+
+  result = g_string_new (NULL);
+
+  next_element = first_element;
+
+  while (TRUE)
+    {
+      const gchar *element;
+      const gchar *start;
+      const gchar *end;
+
+      if (next_element)
+	{
+	  element = next_element;
+	  next_element = va_arg (args, gchar *);
+	}
+      else
+	break;
+
+      start = element;
+      
+      if (is_first)
+	is_first = FALSE;
+      else if (separator_len)
+	{
+	  while (start &&
+		 strncmp (start, separator, separator_len) == 0)
+	    start += separator_len;
+	}
+
+      end = start + strlen (start);
+      
+      if (next_element && separator_len)
+	{
+	  while (end > start + separator_len &&
+		 strncmp (end - separator_len, separator, separator_len) == 0)
+	    end -= separator_len;
+	}
+
+      if (end > start)
+	{
+	  if (result->len > 0)
+	    g_string_append (result, separator);
+
+	  g_string_append_len (result, start, end - start);
+	}
+    }
+  
+  return g_string_free (result, FALSE);
+}
+
+/**
+ * g_build_path:
+ * @separator: a string used to separator the elements of the path.
+ * @first_element: the first element in the path
+ * @Varargs: remaining elements in path
+ * 
+ * Creates a path from a series of elements using @separator as the
+ * separator between elements. At the boundary between two elements,
+ * any trailing occurrences of separator in the first element, or
+ * leading occurrences of separator in the second element are removed
+ * and exactly one copy of the separator is inserted.
+ * 
+ * Return value: a newly-allocated string that must be freed with g_free().
+ **/
+gchar *
+g_build_path (const gchar *separator,
+	      const gchar *first_element,
+	      ...)
+{
+  gchar *str;
+  va_list args;
+
+  g_return_val_if_fail (separator != NULL, NULL);
+
+  va_start (args, first_element);
+  str = g_build_pathv (separator, first_element, args);
+  va_end (args);
+
+  return str;
+}
+
+/**
+ * g_build_filename:
+ * @first_element: the first element in the path
+ * @Varargs: remaining elements in path
+ * 
+ * Creates a filename from a series of elements using the correct
+ * separator for filenames. This function behaves identically
+ * to <literal>g_build_path (G_DIR_SEPARATOR_S, first_element, ....)</literal>.
+ *
+ * No attempt is made to force the resulting filename to be an absolute
+ * path. If the first element is a relative path, the result will
+ * be a relative path. 
+ * 
+ * Return value: a newly-allocated string that must be freed with g_free().
+ **/
+gchar *
+g_build_filename (const gchar *first_element, 
+		  ...)
+{
+  gchar *str;
+  va_list args;
+
+  va_start (args, first_element);
+  str = g_build_pathv (G_DIR_SEPARATOR_S, first_element, args);
+  va_end (args);
+
+  return str;
+}
